/************************************************************************************
*************************************************************************************
*main.c: 
*SECE Explorer Application. Creates service discovery thread and communicates with the discovered
*services to the secs module
*
*Copyright (c) 2011 by Manu Dhundi. All rights reserved.
*
*************************************************************************************
************************************************************************************/
#include <stdio.h>			/* fprintf() */
#include <string.h>		/* strcmp(), strcpy() */
#include <stdlib.h>		/* exit() */

#include <semaphore.h>
#include "pthread.h"   				/* Thread Functions */
 
#include "sece_config.h" 					/* For any config parameters */
#include "secs.h"											/* secsSendList() and associated data structures*/ 
#include "avahidiscover.h"				/* serviceDiscoveryStruct, sdwInit() */

sem_t gServiceDiscWaitSem;		/* wait for service to be discovered.*/
sem_t gServerCommWaitSem; /* wait to finish communication with the server.*/
int gServicesCount = 0; 						/* Number of Services Discovered */
serviceDiscoveryStruct* pServices = NULL; /*Current ptr to linked list of services*/
serviceDiscoveryStruct* pHeadPtr = NULL;  /*Head ptr to linked list of services*/

serviceInfoStruct gAvailableServices[MAXSERVICES] ; /* All available services stored here */
serviceInfoStruct gCurrentService ; /* Current Service that is discovered is stored here */


/***************************************************************************
* printlist()
*prints the list of services (names) discovered by the client
*
***************************************************************************/
static void printlist(void)
{
	serviceDiscoveryStruct *pTemp = pHeadPtr;
	
	/*Travers through the linked list and print the service name*/
	while(pTemp) {
		fprintf(stderr, "service name is %s\n",pTemp->serviceName);
		pTemp = pTemp->pNext;
	}
}

/***************************************************************************
* copyCurrentServiceForServerComm()
* copy the service discovered currently for sending it to the server
*
***************************************************************************/
static void copyCurrentServiceForServerComm(serviceDiscoveryStruct* service)
{
	strcpy(gCurrentService.serviceName, service->serviceName);
	strcpy(gCurrentService.serviceType, service->serviceType);
	strcpy(gCurrentService.serviceIP, service->serviceIP);
	strcpy(gCurrentService.hostName, service->hostName);
	strcpy(gCurrentService.port, service->port);
	gCurrentService.serviceAddFlag = service->serviceAddFlag;
}

/***************************************************************************
* serviceDiscoveryCallback()
* Will be called by SDW if there is an addition or deletion in list of services
*
***************************************************************************/
int serviceDiscoveryCallback(serviceDiscoveryStruct *serviceDetails)
{
	if(serviceDetails->serviceAddFlag) { /*Service to be added*/
		gServicesCount = gServicesCount+1;
		/*I am keeping 2 pointers for the linked list.
		  Head pointer and the pointer to last node*/
		if(pServices) {
			serviceDiscoveryStruct* temp;
			temp = (struct serviceDiscoveryStruct *) malloc(sizeof(struct serviceDiscoveryStruct));
			pServices->pNext = temp;  /*Link the previous node to this*/
			pServices = temp;
			pServices->pNext = NULL;  /*Since last node*/
		}
		else { /*first node*/
			serviceDiscoveryStruct* temp;
			temp = (struct serviceDiscoveryStruct *) malloc(sizeof(struct serviceDiscoveryStruct));
			pServices = temp;
			pHeadPtr = temp;      /*Since First node*/
			pServices->pNext = NULL;  /*Since the only node*/
		}
		
		/*Now fill data to the node*/
		strcpy(pServices->serviceName, serviceDetails->serviceName);
		strcpy(pServices->serviceType, serviceDetails->serviceType);
		strcpy(pServices->serviceIP, serviceDetails->serviceIP);
		strcpy(pServices->hostName, serviceDetails->hostName);
		strcpy(pServices->port, serviceDetails->port);
		strcpy(pServices->serviceDescription, serviceDetails->serviceDescription);
		pServices->serviceAddFlag = serviceDetails->serviceAddFlag ;
		
		/* Copy the data to indicate the server*/
		copyCurrentServiceForServerComm(pServices);
	}
	else	{ /* Remove the service*/
		int removeFlag = 0; /* Flag to indicate if the service to be removed is present */
		int headNodeFlag = 1; /* Flag to indicate if the node removed is the head node */
		serviceDiscoveryStruct *temp =  pHeadPtr;
		serviceDiscoveryStruct  *temp1 =  pHeadPtr;
		/*temp1 to keep track of previous node of the node to be deleted*/
		
		/*Find the service to be removed*/
		while(temp) {
			if(0 == strcmp(temp->serviceName, serviceDetails->serviceName)) {
				removeFlag = 1;
				break;
			}
			else {
				temp1 = temp; /*Previous node in the list*/
				temp = temp->pNext;
			}
			headNodeFlag = 0; /*Going 2nd time in loop, hence head node is not removed*/
		}
		/*Remove the service*/
		if(removeFlag) {
			temp->serviceAddFlag = 0;
			
				/* Copy the data (before removing from the local list) to indicate the server*/
			copyCurrentServiceForServerComm(temp);
			if(headNodeFlag)  { /* Update the head pointer*/
				pHeadPtr = temp->pNext;
			}
			else {
				temp1->pNext = temp->pNext;
				/* Update the last (current) node pointer */
				if(NULL == temp1->pNext) {
					pServices = temp1;
				}				
			}	
			free(temp);
			gServicesCount = gServicesCount-1;	
		}
		else {
			fprintf(stderr,"Service %s was not in the list\n",serviceDetails->serviceName);
			return 0; /* Return without initiating server communication */
		}
	}
	
	printlist();
	//copyServicesForServerComm();
	
	/*Indicate the start of communication with the server*/
	sem_post(&gServiceDiscWaitSem);
	
	/*Wait till server communication over*/
	sem_wait(&gServerCommWaitSem);
	return 0;
}

/***************************************************************************
* sdwThread()
* Separate thread for device discovery
*
***************************************************************************/
static void* sdwThread(void* threadArg)
{
	int iRetVal = 0;
	/* This thread will continuously poll for services in the network */
	iRetVal = sdwInit(serviceDiscoveryCallback);
	pthread_exit(NULL);
}



/***************************************************************************
* main()
*
*
***************************************************************************/
int main()
{
	int iRetVal = 0;
	pthread_t thread1; /* Service discovery thread */
	
	/*Initialize the semaphores*/
	iRetVal = sem_init(&gServiceDiscWaitSem, 0,0);
	if(-1 == iRetVal)
	{
		fprintf(stderr,"sem_init failed\n");
		exit(1);
	}

	iRetVal = sem_init(&gServerCommWaitSem, 0,0);
	if(-1 == iRetVal)
	{
		fprintf(stderr,"sem_init failed\n");
		exit(1);
	}

	/* SDW init in a new thread
	  Service Discovery is done in a separate thread.
		Will continuously poll for services in the network 
    The callback will indicate the discovery of a new service */
	iRetVal = pthread_create( &thread1, NULL, sdwThread, NULL);

	if(0!=iRetVal)
	{
		fprintf(stderr,"pthread_create failed\n");
		exit(1);
	}
	
	/* SECS init*/
	iRetVal = secsInit();
	if(0!=iRetVal)
	{
		fprintf(stderr,"secsInit failed\n");
		exit(1);
	}

  /*Wait for services to be discovered*/
	while(1)
	{
		sem_wait(&gServiceDiscWaitSem);
		/*If the services are discovered, send them to server*/	
		iRetVal = secsSendServices(&gCurrentService, 1);
		
		/*Notify that comminication with the server is done*/
		sem_post(&gServerCommWaitSem);
		
	}

	pthread_join( thread1, NULL);
	sem_destroy(&gServiceDiscWaitSem);
	sem_destroy(&gServerCommWaitSem);
	return 0;

}


#if 0
/***************************************************************************
* copyServicesForServerComm()
* copy from the discovery structure to server communication structure.
*
***************************************************************************/
static void copyServicesForServerComm(void)
{
	int count = 0;
	serviceDiscoveryStruct *pMovePtr = pHeadPtr;
	while(pMovePtr)	{
		strcpy(gAvailableServices[count].serviceName, pMovePtr->serviceName);
		strcpy(gAvailableServices[count].serviceType, pMovePtr->serviceType);
		strcpy(gAvailableServices[count].serviceIP, pMovePtr->serviceIP);
		strcpy(gAvailableServices[count].hostName, pMovePtr->hostName);
		strcpy(gAvailableServices[count].port, pMovePtr->port);
		//strcpy(gAvailableServices[count].serviceDescription, pServices->serviceDescription);

		/*Make pServices point to next node*/
		pMovePtr = pMovePtr->pNext;
		count ++;
		
		/* Check if the server communication structure array has enough space*/
		if(MAXSERVICES == count){
			fprintf(stderr, "Number of services exceeded max services\n");
			break;
		}
	}
}
#endif
