/************************************************************************************
*************************************************************************************
*http.c: 
*Contains functions necessary to form HTTP request and to parse HTTP response
*Note: Contains only bare minimum functions that are needed by the application
*
*Copyright (c) 2011 by Manu Dhundi. All rights reserved.
*
*************************************************************************************
************************************************************************************/

#include <stdio.h>			/* fprintf() */
#include <stdlib.h>    /* atoi()*/
#include <string.h>	  /* strncpy(), strlen()*/

#define HTTPVER " HTTP/1.1"
#define NEXTLINE "\r\n"
#define COLON ": "


/***************************************************************************
* mystrcopy()
* This returns the last location of the destination. Very handy when u want
* to copy multiple strings to a destination buffer
*
*
***************************************************************************/
static char* mystrcopy(char* dest, char* src)
{
	while(*src){
		*(dest) = *(src);
		src++;
		dest++;
	}
	*dest = '\0';
	return dest;
}


/*****************************************************************
* httpHeaderInit()
* Initialize the HTTP header with Method and Url
*
* returns the pointer to the end of the buffer on success and NULL on failure
*****************************************************************/
char* httpHeaderInit(char* pBuf, char* pHttpMethod, char* pUrl)
{
	char *p; /*Pointer for string manipulations*/
	char *path;
	char *host;
	
	if((pBuf == NULL) || (pHttpMethod == NULL) || (pUrl == NULL)){
		fprintf(stderr, "Error in httpHeaderInit\n");
		return NULL;
	}
	
	/*Assign p to the dest to start filling the dest*/
	p = pBuf;
	
	/*It is assumed that a valid method is passed*/
	p = mystrcopy(p, pHttpMethod);
	
	/*Find for the path in the url*/
	path = strstr(pUrl, "//");
	if(path == NULL){
		fprintf(stderr, "Invalid url\n"); /*Not a HTTP url*/
		return NULL;
	}
	path = path+2;  /*Move to location after '//' */
	host = path;      /*Store the host location in pointer host*/
	path = strstr(path, "/");
	
	/*Find if there is path after the url*/
	if(path == NULL){
		p = mystrcopy(p," /");
	}
	else{
		p = mystrcopy(p,"  ");
		p = mystrcopy(p,path);
	}
	
	/*Add HTTP 1.1*/
	p = mystrcopy(p,HTTPVER);
	p = mystrcopy(p,NEXTLINE);
	
	/*Add Host name*/
	p = mystrcopy(p,"Host: ");
	
	/*Add portion of the url without the path*/
	if(path == NULL){
		p = mystrcopy(p,host);
	}
	else{
		p = strncpy(p, host, (path - host));
		p = p + (path - host);
	}
	/* \r\n*/
	p = mystrcopy(p,NEXTLINE);
	return p;
}


/*****************************************************************
* httpAddField()
* Adds the field and value to already initialized header
*
*  returns the pointer to the end of the buffer on success and NULL on failure
*****************************************************************/
char* httpAddField(char* pBuf, char* pHttpFieldName, char* pHttpFieldValue)
{
	char *p; /*Pointer for string manipulations*/
	
	if((pBuf == NULL) || (pHttpFieldName == NULL) || (pHttpFieldValue == NULL)){
		fprintf(stderr, "Error in httpAddField\n");
		return NULL;
	}
	
	/*Assign p to the dest to start filling the dest*/
	p = pBuf;
	
	/*Add Field Name*/
	p = mystrcopy(p, pHttpFieldName);
	
	/* Add Colon and Space*/
	p = mystrcopy(p, COLON);
	
	/* Add Field Value*/
	p = mystrcopy(p, pHttpFieldValue);
	
	p = mystrcopy(p,NEXTLINE);
	return p;		
}

/*****************************************************************
* httpHeaderEnd()
* HTTP header ends with an extra \r\n
*
*  returns the pointer to the end of the buffer on success and NULL on failure
*****************************************************************/
char* httpHeaderEnd(char* pBuf)
{
	char *p = pBuf; /*Pointer for string manipulations*/
	
	if(pBuf == NULL){
		fprintf(stderr, "Error in httpHeaderEnd\n");
		return NULL;
	}
	p = mystrcopy(p,NEXTLINE);
	return p;	
}


/*****************************************************************
* httpParseStatusCode()
* Parses pBuf for httpStatusCode. The status code sent to the calling function
* through pStatusCode
*
*returns 0 on success and -1 on failure.
*****************************************************************/
int httpParseStatusCode(char* pBuf, int* pStatusCode)
{
    char cTemp[5] = {0};
	if(pBuf == NULL){
		fprintf(stderr, "Error in httpParseStatusCode\n");
		return -1;
	}
	/* Move beyond HTTP/1.1 */
	pBuf = pBuf + strlen(HTTPVER);
	
	/*Copy only the 3 digit Status Code */
	strncpy(cTemp, pBuf, 3);
	*pStatusCode = atoi(cTemp);
	return 0;
	
}