/************************************************************************************
*************************************************************************************
*websocket.c: 
*SECE Client uses this to establish communication with SECE Server over Websocket.
*The server sends commands to control devices in the home network over the websocket
*
*Copyright (c) 2011 by Manu Dhundi. All rights reserved.
*
*************************************************************************************
************************************************************************************/

#include <stdio.h>			/* fprintf() */
#include <stdlib.h>		/* exit() */
#include <string.h>		/* strcmp(), strcpy(), strstr(), strncpy()*/

#include <json/json.h>  /* To parse json doc*/

#include <unistd.h>           /* close() */
#include <sys/socket.h>    /* For Socket Calls */
#include <netdb.h>						/* Structure for hostent*/
#include <arpa/inet.h>			/* inet_addr()*/

#include "http.h"               /* To make HTTP related calls*/

#define MAXBUFLEN 2048  /* Send and recevie buffer size*/
#define DESTPORT 80
#define SERVERADDR "http://localhost/"
#define SERVERADDRMAXLEN	256

#define NOSERVER


/*****************************************************************
* sendCommandToService()
*
* sends command to the specified service
*
*****************************************************************/
int sendCommandToService(char* cServiceIP, char* cCommand, char* cPort)
{
	int sockfd;                             /* socket descriptor */
	struct sockaddr_in servAddr; /* service's address information */
	int iBytesSent = 0;
	int iPort = 0;
	
	/*Create a socket for connection with server*/
	if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) == -1){
		fprintf(stderr, "Error in socket creation\n");
		exit(1);
  }
	
	iPort = atoi(cPort);
	/* set service addr*/
	servAddr.sin_family = AF_INET;   			/* host byte order */
	servAddr.sin_port = htons(iPort); /* short, network byte order */
	servAddr.sin_addr.s_addr = inet_addr(cServiceIP); /* fill service IP*/
	memset(servAddr.sin_zero, '\0', sizeof servAddr.sin_zero);
	
	fprintf(stderr, "Sending command to %s\n",cServiceIP);
	/*Send command over UDP*/
	if ((iBytesSent = sendto(sockfd, cCommand, strlen(cCommand), 0,
    (struct sockaddr *)&servAddr, sizeof servAddr)) == -1){
		fprintf(stderr, "Error in sendto\n");
		exit(1);
  }
	return 0;
}

/*****************************************************************
* json_parse()
*
* Parsing the json object
*
*****************************************************************/
void json_parse(json_object * jobj)
{
	enum json_type type;
	char cServiceIP[64] = {0};
	char cCommand[64] = {0};
	char cPort[10] = {0};
	json_object_object_foreach(jobj, key, val) { /*Passing through every array element*/
		type = json_object_get_type(val);
		switch (type) {
			case json_type_boolean:
			case json_type_double:
			case json_type_int:
			case json_type_string: 
				printf(" %s : %s\n",key, json_object_get_string(val));
				if(strcmp("ServiceIP",key) == 0) {
					strcpy(cServiceIP, json_object_get_string(val));
				}
				else if(strcmp("Command",key) == 0) {
					strcpy(cCommand, json_object_get_string(val));
				}
				else if(strcmp("Port",key) == 0) {
					strcpy(cPort, json_object_get_string(val));
				}
				else {
				}
				break;
			case json_type_null:
			case json_type_array:
			case json_type_object:
			default:
				fprintf(stderr, "unexpected json type\n");
			break;
		}
	}
	sendCommandToService(cServiceIP, cCommand, cPort);
}

/*****************************************************************
* parseServerRequest()
*
* Parse the request from sever which is in form of JSON doc
*
*****************************************************************/
static int parseServerRequest(char* request)
{
	json_object * jobj = NULL;
#ifdef NOSERVER
	/*I will parse the JSON file stored locally until server is stored*/
	FILE *pFilefd ;
	char jsonBuf[100] = {0};
	
	pFilefd = fopen ( "serverreq.json" , "r" );
	fread(jsonBuf, 1, 100, pFilefd);
	printf("jsonBuf: %s\n",jsonBuf);
	
	jobj = json_tokener_parse(jsonBuf);
#else
	jobj = json_tokener_parse(request);
#endif
	json_parse(jobj);
	return 0;
}

/*****************************************************************
* main()
*
*
*
*****************************************************************/
int main(int argc, char *argv[])
{
	int iRetVal = 0;
	int sockfd;                             /* socket descriptor */
	struct sockaddr_in dest_addr; /* connector's address information */
	struct hostent *he;
	int numbytes = 0;  /* To count number of bytes sent and received */
	char cSendBuf[MAXBUFLEN] = {0};
	char cReceiveBuf[MAXBUFLEN] = {0};
	int iHttpResponse = 0;
	
	char cHostname[SERVERADDRMAXLEN] = {0};
	
	/*For read timeout*/
	struct timeval tv;
	fd_set readfds;

	char *p; /*Pointers for string manipulations*/
	char *p1;

#ifdef NOSERVER
	FILE *pFilefd ;
	char cServerResponseBuf[MAXBUFLEN] = {0};
#endif
	
	/* Create Http header*/
	p = httpHeaderInit(cSendBuf, "GET", SERVERADDR);
	
	/* Add Websocket Fields*/
	p = httpAddField(p, "Upgrade", "WebSocket");
	p = httpAddField(p, "Connection", "Upgrade");
	p = httpAddField(p, "Sec-WebSocket-Version", "6");
	
	/*All fields added*/
	httpHeaderEnd(p);
	
	fprintf(stderr, "HTTP Header %s\n",cSendBuf );
	
	/*Remove path from url*/
	p = strstr(SERVERADDR, "//");
	p = p +2;
	p1 = strstr(p, "/");
	if(p1){
		strncpy(cHostname, p, (p1 - p));
	}
	else{
		strcpy(cHostname,p);
	}
	
	/*Create a socket for connection with server*/
	if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1){
		fprintf(stderr, "Error in socket creation\n");
		exit(1);
  }
	
	/*Get server IP*/
	if((he = gethostbyname(cHostname)) == NULL) { /* get the host info */
		fprintf(stderr, "Error in gethostbyname\n");
		close(sockfd);
		exit(1);
    }
	
	dest_addr.sin_family = AF_INET;   /* host byte order */
	dest_addr.sin_port = htons(DESTPORT); /* short, network byte order */
	dest_addr.sin_addr = *((struct in_addr *)he->h_addr);
	memset(dest_addr.sin_zero, '\0', sizeof dest_addr.sin_zero);

	/* Connect to server*/
	if (connect(sockfd, (struct sockaddr *)&dest_addr, sizeof dest_addr) == -1){
		fprintf(stderr, "Error in connect\n");
		close(sockfd);
		exit(1);
	}
	 
	/*Send HTTP request to server. Request for websocket establishment*/
	if ((numbytes = send(sockfd, cSendBuf, strlen(cSendBuf), 0)) == -1){
		fprintf(stderr, "Error in send\n");
		close(sockfd);
		exit(1);
	}
	
	/*Set the timeout value for read*/
	tv.tv_sec = 2;
	tv.tv_usec = 0; /*Should be assigned to 0 to avoid errors*/
	
	FD_ZERO(&readfds);
	FD_SET(sockfd, &readfds);
	
	/*Wait for read socket to be ready*/
	select(sockfd+1, &readfds, NULL, NULL, &tv);
	
	if (FD_ISSET(sockfd, &readfds)){
		if ((numbytes = recv(sockfd, cReceiveBuf, MAXBUFLEN-1 , 0)) == -1){
			fprintf(stderr, "Error in recv\n");
			close(sockfd);
			exit(1);
		}
  }
	
#ifdef NOSERVER
	/*I will read a dummy response stored locally until server is ready*/
	pFilefd = fopen ("HttpServerResponse.txt" , "r" );
	fread(cServerResponseBuf, 1, MAXBUFLEN, pFilefd);
	fprintf(stderr, "cServerResponseBuf: %s\n",cServerResponseBuf);
	iRetVal = httpParseStatusCode(cServerResponseBuf, &iHttpResponse);
	fprintf(stderr, "Http response is %d\n",iHttpResponse);
#else
	iRetVal = httpParseStatusCode(cReceiveBuf, &iHttpResponse);
#endif

	if(101 == iHttpResponse) /*Check if the server has accepted the websocket mode of comm.*/
	{
		while(1){
			fprintf(stderr, "Waiting on websocket\n");
			
			/*It is important to set the timeout value again. For every timeout, this value decreases to 0*/
			tv.tv_sec = 5;
			tv.tv_usec = 0; /*Should be assigned to 0 to avoid errors*/
			
			FD_ZERO(&readfds);
			FD_SET(sockfd, &readfds);
			
			/*Setting the timeout parameter as NULL means block indefinitely.*/
			select(sockfd+1, &readfds, NULL, NULL, &tv);
		
			if (FD_ISSET(sockfd, &readfds)){
				if ((numbytes = recv(sockfd, cReceiveBuf, MAXBUFLEN-1 , 0)) == -1){
					fprintf(stderr, "Error in recv\n");
					close(sockfd);
					exit(1);
				}
				if(numbytes == 0){
					fprintf(stderr, "Server Closed the connection\n");
					break;
				}
		  }
		}
		
		/*Need to conform whether cReceiveBuf[numbytes-1] = 0*/
		cReceiveBuf[numbytes] = 0;
		parseServerRequest(cReceiveBuf);
	}
	else{
		fprintf(stderr, "Server refuses Websocket connection");
		close(sockfd);
		exit(1);
	}
	close(sockfd);
	return 0;
}